/*
// UNIT TESTS OF THE BOT
                               
                                                   brake  M1   brake  M2   brake  M3    LACT
INPUT: FULL BRAKE Command String                = ~+0001,+0000,+0001,+0000,+0001,+0000,+0000
       Release Brakes Command String            = ~+0000,+0000,+0000,+0000,+0000,+0000,+0000
       Full Motor Forward String                = ~+0000,+0100,+0000,+0100,+0000,+0000,+0000
       Full Motor Backward String               = ~+0000,-0100,+0000,-0100,+0000,+0000,+0000

Turn Right                                      = ~+0000,+0100,+0000,-0100,+0000,+0000,+0000
Turn Left                                       = ~+0000,-0100,+0000,+0100,+0000,+0000,+0000
       
       Left Motor Command String                = ~+0000,+0100,+0000,+0000,+0000,+0000,+0000
       Right Motor Command String               = ~+0000,+0000,+0000,+0100,+0000,+0000,+0000
       Flipper forwards MotorCommand String     = ~+0000,+0000,+0000,+0000,+0000,+0100,+0000
       Flipper backwards MotorCommand String    = ~+0000,+0000,+0000,+0000,+0000,-0100,+0000
       LACT forwards MotorCommand String        = ~+0000,+0000,+0000,+0000,+0000,+0000,+3200
       LACT backwards MotorCommand String       = ~+0000,+0000,+0000,+0000,+0000,+0000,-3200


OUTPUT: ir sensor array in cm : (fl, fr, l, r, bl, br) from front left to back right


--------- NEW TARGET Motor Values -----------
target_Motor1_Command: 0
target_Motor2_Command: 0
target_FlipperMotor_Command: 0
target_LACT_Command: 0
---------------------------------------
blip - 60 milliseconds elapsed --- ( 97, 68, 129, 204, 190, 83)

*/
//----- LACT Hookups -----
#include <SoftwareSerial.h>
#define rxPin 22  // pin 3 connects to smcSerial TX  (not used in this example)
#define txPin 23  // pin 4 connects to smcSerial RX
SoftwareSerial smcSerial = SoftwareSerial(rxPin, txPin);


//----- Motor Hookups -----
#define TRED_R_MOTOR_SPEED_PIN       2
#define TRED_R_MOTOR_BRAKE_PIN       3
#define TRED_R_MOTOR_DIRECTION_PIN   25

#define TRED_L_MOTOR_SPEED_PIN       4
#define TRED_L_MOTOR_BRAKE_PIN       5
#define TRED_L_MOTOR_DIRECTION_PIN   27

#define FLIPPER_MOTOR_SPEED_PIN       6
#define FLIPPER_MOTOR_BRAKE_PIN       7
#define FLIPPER_MOTOR_DIRECTION_PIN   29

#define RAMP_SPEED_SLOW 20
#define RAMP_SPEED_MED 10
#define RAMP_SPEED_FAST 1

//----- DistanceSaftey Defines -----
#define FORWARD_SAFE_DISTANCE     25
#define BACKWARD_SAFE_DISTANCE    25

//----- KeepAlive Variables -----
#define KEEP_ALIVE_TIMEOUT_1      10000 
#define KEEP_ALIVE_TIMEOUT_2      6 ////roughly every 2 seconds if we use 20 but lets try 2
unsigned long start, finished, elapsed;
int keepAlive_1 = KEEP_ALIVE_TIMEOUT_1;
int keepAlive_2 = KEEP_ALIVE_TIMEOUT_2;

int SAFTEY_SHUTDOWN_FORWARD = 0;
int SAFTEY_SHUTDOWN_BACKWARD = 0;

//----- SerialComms Variables -----
char targetCommandString[] = {'0','0','0','0','0'};

int target_Motor1_Command = 0;
int target_Motor1_brake_command = 0;
int target_Motor2_Command = 0;
int target_Motor2_brake_command = 0;
int target_FlipperMotor_Command = 0;
int target_FlipperMotor_brake_command = 0;
int target_LACT_Command = 0;

//----- IR BLOCK -----
#include <SharpIR.h>

#define IR_FRONT_LEFT         A5
#define IR_FRONT_RIGHT        A4
#define IR_LEFT               A3
#define IR_RIGHT              A2
#define IR_BACK_LEFT          A1
#define IR_BACK_RIGHT         A0

#define LONG_RANGE_IR_SENSOR  20150
#define MED_RANGE_IR_SENSOR   1080

SharpIR IR_FL(IR_FRONT_LEFT, LONG_RANGE_IR_SENSOR);
SharpIR IR_FR(IR_FRONT_RIGHT, LONG_RANGE_IR_SENSOR);
SharpIR IR_L(IR_LEFT, MED_RANGE_IR_SENSOR);
SharpIR IR_R(IR_RIGHT, MED_RANGE_IR_SENSOR);
SharpIR IR_BL(IR_BACK_LEFT, LONG_RANGE_IR_SENSOR);
SharpIR IR_BR(IR_BACK_RIGHT, LONG_RANGE_IR_SENSOR);
//--------------------


void setup() {
  // put your setup code here, to run once:
  Serial.begin( 250000 );

  //----- LACT SETUP -----
  smcSerial.begin(19200);
  delay(5);
  smcSerial.write(0xAA);  // send baud-indicator byte
  exitSafeStart();  // clear the safe-start violation and let the motor run

 
  Serial.println("iBot v1.0 by OrbitusRobotics");
  Serial.println("BEGIN STARTUP SEQUENCE");
  //---- MotorPin SETUP ----
  pinMode( TRED_R_MOTOR_SPEED_PIN, OUTPUT);   
  pinMode( TRED_R_MOTOR_BRAKE_PIN, OUTPUT);
  pinMode( TRED_R_MOTOR_DIRECTION_PIN, OUTPUT);   

  pinMode( TRED_L_MOTOR_SPEED_PIN, OUTPUT);   
  pinMode( TRED_L_MOTOR_BRAKE_PIN, OUTPUT);
  pinMode( TRED_L_MOTOR_DIRECTION_PIN, OUTPUT);   

  pinMode( FLIPPER_MOTOR_SPEED_PIN, OUTPUT);   
  pinMode( FLIPPER_MOTOR_BRAKE_PIN, OUTPUT);
  pinMode( FLIPPER_MOTOR_DIRECTION_PIN, OUTPUT);   

  //--------
  //ZERO THE SPEED
  analogWrite(TRED_L_MOTOR_SPEED_PIN, 255);
  analogWrite(TRED_R_MOTOR_SPEED_PIN, 255);
  //THEN RELEASE THE BRAKE
  releaseBrake();
  //--------

  
  
  //---- IR BLOCK SETUP ----
  pinMode( IR_FRONT_LEFT, INPUT);
  pinMode( IR_FRONT_RIGHT, INPUT);
  pinMode( IR_LEFT, INPUT);
  pinMode( IR_RIGHT, INPUT);
  pinMode( IR_BACK_LEFT, INPUT);
  pinMode( IR_BACK_RIGHT, INPUT);
  //------------------------

  Serial.println("STARTUP SEQUENCE COMPLETE");
  Serial.println("iBot v1.0 ONLINE");
 
  start=millis();

}

void loop() {
  // put your main code here, to run repeatedly:
    keepAlive_SAFTEYSTOP();
    
    if(Serial.available() > 0) {
 
      Serial.print("Parsing Serial and resetting keepAlive\n");
      resetKeepAlive();
      
      byte b =  Serial.read();
      
      if (b == 126)
      {
        char charIn = 0;

        targetCommandString[0] = Serial.read();
        targetCommandString[1] = Serial.read();
        targetCommandString[2] = Serial.read();
        targetCommandString[3] = Serial.read();
        targetCommandString[4] = Serial.read();

        target_Motor1_brake_command = atoi(targetCommandString);
        Serial.print("target_Motor1_brake_command ");
        Serial.println(target_Motor1_brake_command );

        Serial.read(); //READ THE COMMA SEPARATING THINGS!!!

  //M1 - LEFT TRED        
        targetCommandString[0] = Serial.read();
        targetCommandString[1] = Serial.read();
        targetCommandString[2] = Serial.read();
        targetCommandString[3] = Serial.read();
        targetCommandString[4] = Serial.read();

        Serial.print("TargetCommandString M1:");
        Serial.println( targetCommandString );
        target_Motor1_Command = atoi(targetCommandString);

        Serial.read(); //READ THE COMMA SEPARATING THINGS!!!
  //M2 - RIGHT TRED
        targetCommandString[0] = Serial.read();
        targetCommandString[1] = Serial.read();
        targetCommandString[2] = Serial.read();
        targetCommandString[3] = Serial.read();
        targetCommandString[4] = Serial.read();

        target_Motor2_brake_command = atoi(targetCommandString);

        Serial.read(); //READ THE COMMA SEPARATING THINGS!!!
        
        Serial.print("target_Motor2_brake_command ");
        Serial.println(target_Motor2_brake_command );
        
        targetCommandString[0] = Serial.read();
        targetCommandString[1] = Serial.read();
        targetCommandString[2] = Serial.read();
        targetCommandString[3] = Serial.read();
        targetCommandString[4] = Serial.read();

        Serial.print("TargetCommandString M2:");
        Serial.println( targetCommandString );
        
        target_Motor2_Command = atoi(targetCommandString);

        Serial.read(); //READ THE COMMA SEPARATING THINGS!!!
  //FLIPPER MOTOR
        targetCommandString[0] = Serial.read();
        targetCommandString[1] = Serial.read();
        targetCommandString[2] = Serial.read();
        targetCommandString[3] = Serial.read();
        targetCommandString[4] = Serial.read();

        target_FlipperMotor_brake_command = atoi(targetCommandString);
        
        Serial.read(); //READ THE COMMA SEPARATING THINGS!!!
        
        Serial.print("target_FlipperMotor_brake_command ");
        Serial.println(target_FlipperMotor_brake_command );
        
        targetCommandString[0] = Serial.read();
        targetCommandString[1] = Serial.read();
        targetCommandString[2] = Serial.read();
        targetCommandString[3] = Serial.read();
        targetCommandString[4] = Serial.read();

        Serial.print("TargetCommandString FLipper:");
        Serial.println( targetCommandString );
        target_FlipperMotor_Command = atoi(targetCommandString);

        Serial.read(); //READ THE COMMA SEPARATING THINGS!!!

  //LACT
        targetCommandString[0] = Serial.read();
        targetCommandString[1] = Serial.read();
        targetCommandString[2] = Serial.read();
        targetCommandString[3] = Serial.read();
        targetCommandString[4] = Serial.read();
  
        Serial.read(); //READ THE COMMA SEPARATING THINGS!!!
        //Serial.print("TargetCommandString LACT:");
        //Serial.print( targetCommandString );
        target_LACT_Command = atoi(targetCommandString);

        
  
        setMotorCommand(target_Motor1_brake_command, target_Motor1_Command, target_Motor2_brake_command, target_Motor2_Command, target_FlipperMotor_brake_command, target_FlipperMotor_Command, target_LACT_Command);
        
        //irReadOut();
        
        Serial.println("---------------------------------------");
        
      }
   }
}


int irReadOut_SAFTEYSTOP()
{
  /* Check for obstacles 
  Serial.print( "( " );
  Serial.print( IR_FL.distance() );
  Serial.print( ", " );
  Serial.print( IR_FR.distance() );
  Serial.print( ", " );
  Serial.print( IR_L.distance() );
  Serial.print( ", " );
  Serial.print( IR_R.distance() );
  Serial.print( ", " );
  Serial.print( IR_BL.distance() );
  Serial.print( ", " );
  Serial.print( IR_BR.distance() );
  Serial.print( ")\n" );
  */
  //FORWARD SAFTEY CHECKS
  if ( IR_FL.distance() < FORWARD_SAFE_DISTANCE ||
       IR_FR.distance() < FORWARD_SAFE_DISTANCE)
  {
    Serial.println( "SAFTEY SHUTDOWN ERROR! OBSTACLE IS BLOCKING FRONT!" );
    SAFTEY_SHUTDOWN_FORWARD = 1;
  }
  else
  {
    SAFTEY_SHUTDOWN_FORWARD = 0;
  }
  
  //BACKWARD SAFTEY CHECKS
  if ( IR_BL.distance() < BACKWARD_SAFE_DISTANCE  ||
       IR_BR.distance() < BACKWARD_SAFE_DISTANCE )
  {
    Serial.println( "SAFTEY SHUTDOWN ERROR! OBSTACLE IS BLOCKING BACK!" );
    SAFTEY_SHUTDOWN_BACKWARD = 1;
  }
  else
  {
    SAFTEY_SHUTDOWN_BACKWARD = 0;
  }
  
  
  return 1;
}

//------ KEEP ALIVE FUNCTIONS ------

void keepAlive_SAFTEYSTOP()
{
  keepAlive_1--;
    if (keepAlive_1 <= 0)
    {
      //MAKE SURE TO SET MOTORS BACK TO ZERO AFTER THE TIMEOUT IN CASE OF FAILURE AT FULL SPEED
      keepAlive_1 = KEEP_ALIVE_TIMEOUT_1;
      keepAlive_2--;

      //Serial.print("blip - ");
      //irReadOut();
      //Serial.print(elapsed);
      //Serial.print(" milliseconds elapsed --- ");
      
      //Serial.println(keepAlive_2);
      irReadOut_SAFTEYSTOP();
    }
    if (keepAlive_2 <= 0)
    {
      finished=millis();
      elapsed=finished-start;
      start=millis();
      Serial.print("KEEPALIVE HAS EXPIRED!!! SHUTTING DOWN CONNECTION NOW -- ");
      Serial.print(elapsed);
      Serial.println(" milliseconds");
      keepAlive_2 = KEEP_ALIVE_TIMEOUT_2;
      
      setMotorCommand(0,+0000, 0,+0000 ,0,+0000, +0000);
    }
}

void resetKeepAlive()
{
  keepAlive_1 = KEEP_ALIVE_TIMEOUT_1;
  keepAlive_2 = KEEP_ALIVE_TIMEOUT_2;
}

//------ MOTOR FUNCTIONS ------

void setMotorCommand(int target_Motor1_Brake_command, int target_Motor1_Command, int target_Motor2_Brake_command, int target_Motor2_Command, int target_FlipperMotor_Brake_command, int target_FlipperMotor_Command, int target_LACT_Command)
{
  Serial.println("--------- NEW TARGET Motor Values -----------");
  Serial.print("target_Motor1_Command: ");
  Serial.print(target_Motor1_Brake_command);
  Serial.print(":");
  Serial.println(target_Motor1_Command);
  Serial.print("target_Motor2_Command: ");
  Serial.print(target_Motor2_Brake_command);
  Serial.print(":");
  Serial.println(target_Motor2_Command);      
  Serial.print("target_FlipperMotor_Command: ");
  Serial.print(target_FlipperMotor_Brake_command);
  Serial.print(":");
  Serial.println(target_FlipperMotor_Command);      
  Serial.print("target_LACT_Command: ");
  Serial.println(target_LACT_Command);      
  Serial.println("---------------------------------------");

  //-----------------------------------
  //Set Brake Commands
  if (target_Motor1_Brake_command == 0)
    digitalWrite(TRED_L_MOTOR_BRAKE_PIN, HIGH);
  else
    digitalWrite(TRED_L_MOTOR_BRAKE_PIN, LOW);
        
  if (target_Motor2_Brake_command == 0)
    digitalWrite(TRED_R_MOTOR_BRAKE_PIN, HIGH);
  else
    digitalWrite(TRED_R_MOTOR_BRAKE_PIN, LOW);
        
  if (target_FlipperMotor_Brake_command == 0)
    digitalWrite(FLIPPER_MOTOR_BRAKE_PIN, HIGH);
  else
    digitalWrite(FLIPPER_MOTOR_BRAKE_PIN, LOW);
        
  //-----------------------------------
  //Set Motor Directions
  if (target_Motor1_Command < 0)
  {
    motor1backward(); //SWITCH ME TO BACKWARDS: refactor
  }
  else
    motor1forward(); //SWITCH ME TO FORWARDS: refactor

  if (target_Motor2_Command < 0)
  {
    motor2backward();//SWITCH ME TO BACKWARDS: refactor
  }
  else
    motor2forward();//SWITCH ME TO FORWARDS: refactor

  if (target_FlipperMotor_Command < 0)
    flipperMotor_backward();
  else
    flipperMotor_forward();  

    //Serial.print("FLIPPERMOTORCOMMAND: ********* ");
    //Serial.print(target_FlipperMotor_Command);
    //Serial.println("********* ");
    
  //-----------------------------------
  //Send the Speed Control Signal

  if ((target_Motor1_Command < 0 && target_Motor2_Command >= 0) ||
      (target_Motor1_Command < 0 && target_Motor2_Command >= 0))
  {
    //isRotating ... allow rotations to happen 
    analogWrite(TRED_L_MOTOR_SPEED_PIN, 255-abs(target_Motor1_Command));
    analogWrite(TRED_R_MOTOR_SPEED_PIN, 255-abs(target_Motor2_Command));

  }
  else
  {
    //Not in a rotation... block any of the treds from crashing into something
    if ((SAFTEY_SHUTDOWN_BACKWARD == 0 && target_Motor1_Command < 0) ||
      (SAFTEY_SHUTDOWN_FORWARD == 0 && target_Motor1_Command >= 0))
      analogWrite(TRED_L_MOTOR_SPEED_PIN, 255-abs(target_Motor1_Command));
  else
      analogWrite(TRED_L_MOTOR_SPEED_PIN, 255);

  if ((SAFTEY_SHUTDOWN_BACKWARD == 0 && target_Motor2_Command < 0) ||
      (SAFTEY_SHUTDOWN_FORWARD == 0 && target_Motor2_Command >= 0))
      analogWrite(TRED_R_MOTOR_SPEED_PIN, 255-abs(target_Motor2_Command));
  else
      analogWrite(TRED_R_MOTOR_SPEED_PIN, 255);

  }
  
  
  analogWrite(FLIPPER_MOTOR_SPEED_PIN, 255-abs(target_FlipperMotor_Command));
  setMotorSpeed(target_LACT_Command);
  //-----------------------------------
}

void releaseBrake()
{
  Serial.println( "***** BRAKE IS OFF!!! *****" );
  digitalWrite(TRED_L_MOTOR_BRAKE_PIN, HIGH);
  digitalWrite(TRED_R_MOTOR_BRAKE_PIN, HIGH);
}


void brake()
{
  Serial.println( "***** BRAKE IS ON!!! *****" );
  digitalWrite(TRED_L_MOTOR_BRAKE_PIN, LOW);
  digitalWrite(TRED_R_MOTOR_BRAKE_PIN, LOW);
}


void backward()
{
  Serial.println( "BACKWARD" );
  digitalWrite(TRED_L_MOTOR_DIRECTION_PIN, HIGH); // (towards back direction)
  digitalWrite(TRED_R_MOTOR_DIRECTION_PIN, LOW); // (towards back direction)
}


void motor1backward()
{
  Serial.println( "M1 FORWARD" );
  
  digitalWrite(TRED_L_MOTOR_DIRECTION_PIN, LOW); // (towards back direction)
}


void motor1forward()
{
  Serial.println( "M1 BACKWARD" );
  
  digitalWrite(TRED_L_MOTOR_DIRECTION_PIN, HIGH); // (towards front direction)
}


void motor2backward()
{
  Serial.println( "M2 FORWARD" );
  
  digitalWrite(TRED_R_MOTOR_DIRECTION_PIN, HIGH); // (towards back direction)
}


void motor2forward()
{
  Serial.println( "M2 BACKWARD" );
  
  digitalWrite(TRED_R_MOTOR_DIRECTION_PIN, LOW); // (towards front direction)
}

void flipperMotor_forward()
{
  Serial.println( "FLIPPER FORWARD" );
  
  digitalWrite(FLIPPER_MOTOR_DIRECTION_PIN, LOW); // (towards front direction)
}

void flipperMotor_backward()
{
  Serial.println( "FLIPPER BACKWARD" );
  
  digitalWrite(FLIPPER_MOTOR_DIRECTION_PIN, HIGH); // (towards back direction)
}


//------ LACT Functions -----

void exitSafeStart()
{
  smcSerial.write(0x83);
}
 
// speed should be a number from -3200 to 3200
void setMotorSpeed(int speed)
{
  if (speed < 0)
  {
    smcSerial.write(0x86);  // motor reverse command
    speed = -speed;  // make speed positive
  }
  else
  {
    smcSerial.write(0x85);  // motor forward command
  }
  smcSerial.write(speed & 0x1F);
  smcSerial.write(speed >> 5);
}
 
